%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Authors: A. Ali and T. Y. Al-Naffouri                                           
% Last Modified: June, 2016
%
% If you use this code or any (modified) part of it in any publication, please cite 
% the paper: A. Ali, M. Masood, M. S. Sohail, S. Al-Ghadhban, and T. Y. Al-Naffouri, Narrowband Interference Mitigation in SC-FDMA Using Bayesian Sparse Recovery, IEEE Transactions on Signal Processing.
%
% Copyright (c) A. Ali, M. Masood, M. S. Sohail, S. Al-Ghadhban, and T. Y. Al-Naffouri
%
% Contact Persons: A. Ali and T. Y. Al-Naffouri
% E-mail: anumali@utexas.edu, tareq.alnaffouri@kaust.edu.sa
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
clc;
clear all;
close all;
%
j=sqrt(-1);
%
M=16; % QAM Constellation Order
N=512; % Number of subcarriers in SC-FDMA system
L=N/4; % Length of the channel (or cyclic prefix)
K=log2(M); % The number of bits
U=2; % Number of Users in SC-FDMA
P=N/U; % Number of sub-carriers per user
max_no_NBI=4; % Maximum number of NBI tones per symbol
%
iter=2e3; % The number of iterations 
%
SIR_dB=-10; % Signal to Interference Ratio
EbNo_dB=[5:2.5:25]; % The Eb / No Ratio
TS=64; % The number of reserved tones per user
%
min_no_errors=2e3; % Minimum number of errors (a simulation parameter)
%
SIR=10^(SIR_dB/10);
SNR_dB=EbNo_dB+10*log10(K);
SNR_vec=10.^(SNR_dB/10);
len_SNR=length(SNR_vec);
%
sgnl_e=mean(abs(qammod(0:M-1,M,0,'gray')).^2); % QAM Constellation Energy
sgnl_e_dB=10*log10(sgnl_e);
%
BER=zeros(7,len_SNR);
RT=zeros(5,len_SNR);
%
FN=1/sqrt(N)*dftmtx(N); % DFT matrix of size N
FHN=FN'; % IDFT matrix of size N
%
FP=1/sqrt(P)*dftmtx(P); % DFT matrix of size P
FHP=FP'; % IDFT matrix of size P
%
Imtx=eye(N);
Pmtx1=Imtx(:,1:U:N); % The selection matrix of user 1
PHmtx1=Pmtx1';
%
Pmtx2=Imtx(:,2:U:N); % The selection matrix of user 2
PHmtx2=Pmtx2';
%
s_rate=sum(1:max_no_NBI)/max_no_NBI/N; % Sparsity Rate
epsilon=1e-3; % A parameter for CVX based ell-1 recovery
%
for jj=1:len_SNR
    SNR=SNR_vec(jj);
    %
    % Variables to store the BER for different cases
    BER_no_NBI1_sum=0; 
    BER_NBI1_sum=0;
    BER_SABMP1_sum=0;
    BER_l11_sum=0;
    BER_FISTA_sum=0;
    BER_SBL_sum=0;
    %
    NE_no_NBI1_sum=0;
    % Variables to store the runtime for different cases
    RT_SABMP1_sum=0;
    RT_l11_sum=0;
    RT_FISTA_sum=0;
    RT_SBL_sum=0;
    %
    iter_run=iter;
    %
    for ii=1:iter
        %==================================================================
        % Common Parameters
        %==================================================================
        [h1,h_cap1,H1,D1]=channel_make(L,N,FN,FHN); % Channel for user 1
        [h2,h_cap2,H2,D2]=channel_make(L,N,FN,FHN); % Channel for user 2
        %
        Mod_mtx1=D1*Pmtx1*FP; % An intermediate variable for ease
        Mod_mtx2=D2*Pmtx2*FP;
        %
        noise_p1=(randn(N,1)+j*randn(N,1)); % Non-Normalized noise vector
        %
        no_of_NBI=randi([1,max_no_NBI],1); % Number of NBI sources (from 1 to 4) 
        NBI_loc= randi([1,N],[no_of_NBI,1]);% + -0.5 +rand(no_of_NBI,1); 
        % There is no grid offset is in this code
        % randi portion generates integers and rand portion perturbs those
        % integers to put the NBI off the grid
        NBI_amp=(randn(no_of_NBI,1)+j*randn(no_of_NBI,1)); % Non-Normalized NBI amplitude vector
        NBI_p1=FN*exp(j*2*pi*(0:N-1)'*NBI_loc'/N)/sqrt(N)*NBI_amp; % Generating the NBI vector
        %
        %==================================================================
        % CS based Transmission and Reception
        %==================================================================
        d1=randi([0,M-1],[P,1]); % Random data digits for user 1
        d_qam1=qammod(d1,M,0,'gray'); % Modulating the data for user 1
        tx_sgnl1=Mod_mtx1*d_qam1; % Transmission signal for user 1
        %
        d2=randi([0,M-1],[P,1]); % Random data digits for user 1
        d_qam2=qammod(d2,M,0,'gray'); % Modulating the data for user 1
        tx_sgnl2=Mod_mtx2*d_qam2; % Transmission signal for user 2
        %
        norm_tx1_tx2=norm(tx_sgnl1+tx_sgnl2); % Signal Power * N
        %
        Z=(norm_tx1_tx2/sqrt(SNR*2*N))*noise_p1; % Normalizing the un-normalized noise based on SNR
        %
        NBI=(norm_tx1_tx2/sqrt(SIR*2*N))*NBI_p1; % Normalizing the un-normalized NBI based on SIR
        %
        Y_no_NBI=tx_sgnl1+tx_sgnl2+Z; % NBI free noisy received signal
        Y_NBI=tx_sgnl1+tx_sgnl2+NBI+Z; % NBI impaired received signal
        %
        A_mtx1=PHmtx1*D1*Pmtx1*FP; % A matrix required for MMSE equalization of user 1
        A_mtx2=PHmtx2*D2*Pmtx2*FP; % A matrix required for MMSE equialization of user 2
        %
        Demod_mtx1=sgnl_e*A_mtx1'/(sgnl_e*(A_mtx1*A_mtx1')+var(Z,1)*eye(P))*PHmtx1; % The equalization matrix for user 1
        %
        Demod_mtx2=sgnl_e*A_mtx2'/(sgnl_e*(A_mtx2*A_mtx2')+var(Z,1)*eye(P))*PHmtx2; % The equalization matrix for user 2
        %
        %==================================================================
        % Deciding the Data Locations and Measurement Locations
        %==================================================================
        % Generating the data and pilot indices for user 1
        index_meas1=randperm(P);
        index_meas1=index_meas1(1:TS);
        index_data1=setdiff(1:P,index_meas1);
        % Generating the data and pilot indices for user 2
        index_meas2=setdiff(1:P,index_meas1); 
        index_meas2=index_meas2(randperm(length(index_meas2)));
        index_meas2=index_meas2(1:TS);
        index_data2=setdiff(1:P,index_meas2);
        %==================================================================
        % No NBI - User 1 (NBI free case)
        %==================================================================
        Y_no_NBI_deperm1=Demod_mtx1*Y_no_NBI; 
        d_demod_no_NBI1=qamdemod(Y_no_NBI_deperm1,M,0,'gray');
        %
        [NE_no_NBI1,BER_no_NBI1]=biterr(d1(index_data1),...
            d_demod_no_NBI1(index_data1),K);
        BER_no_NBI1_sum=BER_no_NBI1_sum+BER_no_NBI1;
        NE_no_NBI1_sum=NE_no_NBI1_sum+NE_no_NBI1;
        %==================================================================
        % NBI - User 1 (NBI impaired case)
        %==================================================================
        Y_NBI_deperm1=Demod_mtx1*Y_NBI;
        d_demod_NBI1=qamdemod(Y_NBI_deperm1,M,0,'gray');
        %
        [~,BER_NBI1]=biterr(d1(index_data1),d_demod_NBI1(index_data1),K);
        BER_NBI1_sum=BER_NBI1_sum+BER_NBI1;
        %==================================================================
        % SABMP based reconstruction
        %==================================================================
        Phi_mtx1=Demod_mtx1(index_meas1,:);
        meas_vec1=Y_NBI_deperm1(index_meas1)-d_qam1(index_meas1);
        %
        norm_mtx1=diag(1./diag(sqrt(Phi_mtx1'*Phi_mtx1)));
        norm_mtx1(isinf(norm_mtx1))=0;
        Phi_mtx1=Phi_mtx1*norm_mtx1;
        %
        Comp_Phi_mtx1=Demod_mtx1;
        %
        tic;
        [NBI1_est,~, ~, ~, ~,~]=nGpFBMP(meas_vec1, ...
            Phi_mtx1, s_rate, 1e-3, 1);
        RT_SABMP1_sum=RT_SABMP1_sum+toc;
        NBI1_est=norm_mtx1*NBI1_est;
        %
        Y_SABMP_deperm1=Y_NBI_deperm1-Comp_Phi_mtx1*NBI1_est;
        d_demod_SABMP1=qamdemod(Y_SABMP_deperm1,M,0,'gray');
        %
        [~,BER_SABMP1]=biterr(d1(index_data1),d_demod_SABMP1(index_data1),K);
        BER_SABMP1_sum=BER_SABMP1_sum+BER_SABMP1;
        %==================================================================
        % l1 - optimization based recovery
        %==================================================================
        tic;
        NBI1_est=cs_alg(meas_vec1, Phi_mtx1, epsilon);
        RT_l11_sum=RT_l11_sum+toc;
        %
        NBI1_est=norm_mtx1*NBI1_est;
        %
        Y_l1_deperm1=Y_NBI_deperm1-Comp_Phi_mtx1*NBI1_est;
        d_demod_l11=qamdemod(Y_l1_deperm1,M,0,'gray');
        %
        [~,BER_l11]=biterr(d1(index_data1),d_demod_l11(index_data1),K);
        BER_l11_sum=BER_l11_sum+BER_l11;
        %==================================================================
        % FISTA based recovery
        %==================================================================
        ProxF=@(x,sigma) sign(x).*(abs(x)-min([abs(x');sigma*ones(size(x'))])');
        GradG=@(x) Phi_mtx1'*(Phi_mtx1*x-meas_vec1);
        options.method='fista';
        tic;
        NBI1_est=perform_fb(zeros(N,1), ProxF, GradG, norm(Phi_mtx1)^2, options);
        RT_FISTA_sum=RT_FISTA_sum+toc;
        %
        NBI1_est=norm_mtx1*NBI1_est;
        %
        Y_FISTA_deperm1=Y_NBI_deperm1-Comp_Phi_mtx1*NBI1_est;
        d_demod_FISTA=qamdemod(Y_FISTA_deperm1,M,0,'gray');
        %
        [~,BER_FISTA]=biterr(d1(index_data1),d_demod_FISTA(index_data1),K);
        BER_FISTA_sum=BER_FISTA_sum+BER_FISTA;
        %==================================================================
        % SBL based recovery
        %==================================================================
        meas_vec1l=[real(meas_vec1);imag(meas_vec1)];
        Phi_mtx1l=[real(Phi_mtx1) -imag(Phi_mtx1); imag(Phi_mtx1) real(Phi_mtx1)];
        tic;
        NBI1_est_str=SparseBayes('Gaussian',Phi_mtx1l,meas_vec1l);
        RT_SBL_sum=RT_SBL_sum+toc;
        NBI1_est=zeros(2*N,1);
        NBI1_est(NBI1_est_str.Relevant) = NBI1_est_str.Value;
        NBI1_est=NBI1_est(1:N)+j*NBI1_est(N+1:end);
        %
        NBI1_est=norm_mtx1*NBI1_est;
        %
        Y_SBL_deperm1=Y_NBI_deperm1-Comp_Phi_mtx1*NBI1_est;
        d_demod_SBL=qamdemod(Y_SBL_deperm1,M,0,'gray');
        %
        [~,BER_SBL]=biterr(d1(index_data1),d_demod_SBL(index_data1),K);
        BER_SBL_sum=BER_SBL_sum+BER_SBL;
        %==================================================================
        % Printing and Counting No of errors
        %==================================================================
        if((NE_no_NBI1_sum)>min_no_errors)
            iter_run=ii;
            floor([jj ii NE_no_NBI1_sum])
            break;
        end
        if(mod(ii,5)==0)
            floor([jj ii NE_no_NBI1_sum])
        end
        %==================================================================
    end
    BER(1,jj)=BER_no_NBI1_sum/iter_run;
    BER(2,jj)=BER_NBI1_sum/iter_run;
    BER(3,jj)=BER_SABMP1_sum/iter_run;
    BER(4,jj)=BER_l11_sum/iter_run;
    BER(5,jj)=BER_FISTA_sum/iter_run;
    BER(6,jj)=BER_SBL_sum/iter_run;
    %
    RT(1,jj)=RT_SABMP1_sum/iter_run;
    RT(2,jj)=RT_l11_sum/iter_run;
    RT(3,jj)=RT_FISTA_sum/iter_run;
    RT(4,jj)=RT_SBL_sum/iter_run;
end
%
figure,semilogy(EbNo_dB,BER(1,:),'gs-'),hold on;
semilogy(EbNo_dB,BER(2,:),'bd-');
semilogy(EbNo_dB,BER(3,:),'ro-');
semilogy(EbNo_dB,BER(4,:),'go-');
semilogy(EbNo_dB,BER(5,:),'mx-');
semilogy(EbNo_dB,BER(6,:),'kp-');
legend('NBI Free','NBI Impaired','Proposed','l1-optimization','FISTA','SBL');
xlabel('Eb/N0'),ylabel('BER')
%
figure,semilogy(EbNo_dB,RT(1,:),'ro-');
hold on, semilogy(EbNo_dB,RT(2,:),'go-');
hold on, semilogy(EbNo_dB,RT(3,:),'mx-');
hold on, semilogy(EbNo_dB,RT(4,:),'kp-');
legend('Proposed','l1-optimization','FISTA','SBL');