% In the name of Allah, The Rahman, The Raheem%
% nGpFBMP:   Function for executing the complex-valued non-Gaussian 
% prior Fast Bayesian Matching Pursuit algorithm.
%
% This function implements the algorithm described in our following paper
% ---------------------------------------------------------------------
% M. Masood and T. Y. Al-Naffouri, "Sparse Reconstruction Using
% Distribution Agnostic Bayesian Matching Pursuit," IEEE Trans. Signal
% Process., vol. 61, no. 21, pp. 5298-5309, Nov. 2013.
% ---------------------------------------------------------------------
% If you use this code or any (modified) part of it in any 
% publication, please cite the above-mentioned paper. 
%
% nGpFBMP will find a high-probability set of sparse solutions "x"
% to the underdetermined system of linear equations
%                           y = Ax + w,
% where w is complex Gaussian, while x may be from any distribution.
%
% SYNTAX:[xmmse,xmmse_Sd,psy_Sd,nu_Sd]=nGpFBMP(y,A,p1,p2,sig2w,D)
%
% Inputs:
% y -       Vector of observations of length M
% A -       M-by-N measurement matrix
% p1 -      A row vector determining prior probability of each element to
%           be active.
% sig2w -   Noise variance
% D -       Maximum number of repeated greedy searches (default 5)
% Outputs:
% xmmse -        approximate MMSE of sparse signal x
% xmmse_star -   significant conditional MMSE estimates of x
% psy_star -     corresponding posterior p(S|y)
% nu_star -      corresponding metric values
% est_err - 	 MMSE estimation error 
% COV_MAT - 	 Error covariance matrix	
%
% Coded by: Mudassir Masood and Tareq Y. Al-Naffouri, KAUST, Saudi Arabia.
% E-mail: mudassir.masood@kaust.edu.sa, tareq.alnaffouri@kaust.edu.sa
% Last change: June 26, 2012
% nGpFBMP version 1.0
% Copyright (c) Mudassir Masood, Tareq Y. Al-Naffouri.


function [xmmse1, xmmse_star, psy_star, nu_star,est_err, COV_MAT] = nGpFBMP(y, A, p1, sig2w, D)


[M, N] = size(A);

% Default algorithmic parameters
P = min(M, 1 + ceil(N*p1 + erfcinv(1e-2)*sqrt(2*N*p1*(1 - p1))));  % search length

%% Repeated greedy search

% Allocate variables for storage

nu = -inf*ones(P,D);	% metrics
xmmse = cell(P,D);
xmmse(:)={zeros(N,1)};


%One time calculations. To be used for calculating E_y and E_psi
col_prod=A'*A;
col_y_prod=A'*y;
Ey1v=transpose(col_y_prod./diag(col_prod));
eta_root=A.*repmat(Ey1v,M,1);

invsig2w=(1/sig2w);
root_const=N*log(1-p1)+(-M/2)*log(2*pi)-invsig2w*norm(y)^2; %constant part of the metrics
c3=log(p1/(1-p1)); %a difference of c3 occurs everytime there is a change in model vector


nuxt_root = root_const + (sum(eta_root.*conj(eta_root),1))*invsig2w + c3; %root node metrics

% Descend one branch at a time
for d = 1:D
    nuxt = nuxt_root; %initializing metrics to the root node
    Eysv=Ey1v; %initializing the root node values of E_y
    cltt=root_const+c3; %initialize with the constant part of the metrics
    for p = 1:P
        Epsv=zeros(p,N);
        Eys_newv=zeros(p+1,N);
        [nustar,nqstar] = max(nuxt);            % find best extension
        while sum(abs(nustar-nu(p,1:d-1)) < 1e-8)   % if same as explored node...
            nuxt(nqstar) = -inf;                    % ... mark extension as redundant
            [nustar, nqstar] = max(nuxt);           % ... and find next best extension
        end
        nstar = mod(nqstar - 1, N) + 1;            % coef index of best extension
        nu(p,d) = nustar;                           % replace worst explored node...
        if (p > 1)
            %preparing variables to compute E_psi (called Epsv in the code)
            q_old=(col_prod(T,nstar))';
            f_old=1-q_old*Eps_oldv(:,nstar);
            c1v=(1/f_old)*(q_old*Eps_oldv(:,1:N)-col_prod(nstar,1:N));
            T=[T nstar];
            
            %compute E_psi
            Epsv(:,1:N)=[Eps_oldv(:,1:N)+Eps_oldv(:,nstar)*c1v; -c1v];
        else
            T=nstar; 
            Epsv=col_prod(nstar,1:N);
        end
        
        Eps_oldv=Epsv; %save Epsv to be used in the next iteration
        Eys_selectedv=Eysv(:,nstar);
        xmmse{p,d}(T) = Eys_selectedv;
        %Eys_selectedv
        %sort(T)
        %preparing variables to compute E_y (called Eys_newv in the code)
        qv=col_prod(T,1:N);
        fv=(1-sum(conj(qv).*Epsv,1));
        c2v=transpose((transpose(1./fv)).*((qv'*Eys_selectedv)-(col_y_prod(1:N))));
        
        %compute E_y
        Eys_newv(:,1:N)=[repmat(Eys_selectedv,1,N)+Epsv(:,1:N).*repmat(c2v,p,1); -c2v];
               
        Eysv=Eys_newv; %store new E_y to be used in next iteration
        cltt=cltt+c3; %a difference of c3 occurs everytime there is a change in model vector
        nuxt=invsig2w*real(sum(conj([repmat(col_y_prod(T),1,N);transpose(col_y_prod)]).*Eys_newv))+cltt; %computing the metrics for next stage
        nuxt(T) = -inf;
    end
end


%% Calculate stuff for statistically significant hypotheses
[~,indx] = sort(nu(:), 'descend');
nu_max = nu(indx(1));
num=length(nu(:));
nu_star = nu(indx(1:num));
psy_star = exp(nu_star - nu_max)/sum(exp(nu_star - nu_max));    % posteriors
xmmse_star = cell(1,num);           % mmse est cond on mixing params

for k = 1:num
    xmmse_star{k} = xmmse{indx(k)};
end

xmmse1 = [xmmse_star{:}]*psy_star;  % approximate mmse estimate

cov1 = zeros(1,num); %vector to hold the trace of the covariance matrix of estimate for each support set
cov2=zeros(1,N);
COV2=cell(1,num);
for k = 1:num
    inds=find(xmmse{indx(k)});
    if(~isempty(inds))
        COV1=((eye(length(inds))/(A(:,inds)'*A(:,inds))*sig2w));
        cov2(inds)=diag(COV1);
        COV2{k}=diag(cov2);
        cov1(k)=(trace(eye(length(inds))/(A(:,inds)'*A(:,inds))*sig2w));% - norm(xmmse_star{k})^2;%  + norm(xmmse_star{k}-xmmse1)^2;
    end
end

est_err = real(cov1*psy_star); %MMSE estimation error

COV_MAT = zeros(N);

for i=1:num
    if(~isempty(COV2{i}))
    COV_MAT = COV_MAT + real(COV2{i}*psy_star(i));
    end
end
COV_MAT = abs(COV_MAT);